# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from dataclasses_json import config


@dataclass
class VersionDTO:
    """Represents a Version Data Transfer Object (DTO)."""

    _version: str = field(metadata=config(field_name="version"))
    """The version of the software."""

    _compile_date: str = field(metadata=config(field_name="compileDate"))
    """The date the software was compiled."""

    _git_commit: str = field(metadata=config(field_name="gitCommit"))
    """The git commit of the software."""

    def version(self) -> str:
        return self._version

    def compile_date(self) -> str:
        return self._compile_date

    def git_commit(self) -> str:
        return self._git_commit
