# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from enum import Enum


class NullOrdering(Enum):
    """A null order used in sorting expressions."""

    NULLS_FIRST: str = "nulls_first"
    """Nulls appear before non-nulls. For ascending order, this means nulls appear at the beginning."""

    NULLS_LAST: str = "nulls_last"
    """Nulls appear after non-nulls. For ascending order, this means nulls appear at the end."""

    def __str__(self) -> str:
        if self == NullOrdering.NULLS_FIRST:
            return "nulls_first"
        if self == NullOrdering.NULLS_LAST:
            return "nulls_last"

        raise ValueError(f"Unexpected null order: {self}")
