# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import ABC, abstractmethod
from typing import List

from gravitino.api.expressions.expression import Expression
from gravitino.api.expressions.sorts.null_ordering import NullOrdering
from gravitino.api.expressions.sorts.sort_direction import SortDirection


class SortOrder(Expression, ABC):
    """Represents a sort order in the public expression API."""

    @abstractmethod
    def expression(self) -> Expression:
        """Returns the sort expression."""
        pass

    @abstractmethod
    def direction(self) -> SortDirection:
        """Returns the sort direction."""
        pass

    @abstractmethod
    def null_ordering(self) -> NullOrdering:
        """Returns the null ordering."""
        pass

    def children(self) -> List[Expression]:
        """Returns the children expressions of this sort order."""
        return [self.expression()]
